/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: geoinfo.c,v 1.4 1996/01/31 20:38:51 deaven Exp deaven $
 */

#include <stdio.h>
#include <string.h>
#include "dmd/tools.h"
#include "dmd/geometry.h"
#include "dmd/element.h"
#include "dmd/memory.h"

char *usage = "usage: %s <options> [geometry]\n"
  "options:\n"
  "\t-c[#]     Print coordination numbers (number of neighbors) [within #]\n"
  "\t-n        Print nearest neighbors\n"
  "\t-I        Print center of mass and moments of inertia\n"
  ;

static int nearest_neighbor(Geometry *g, int i, double *d)
{
	int j, jmin = 0;
	double dd, dij[3], dmin;

	if(g->n < 2) {
	  dmin = 0.0;
	  jmin = 0;
	}
	else {
	  dmin = -1.0;
	  for(j=0; j<(g->n); ++j) {
	    if(j != i) {
	      dd = Gdisp2(g, i, j, dij);
	      if(dmin < 0.0 || dd < dmin) {
		dmin = dd;
		jmin = j;
	      }
	    }
	  }
	}
	*d = sqrt(dmin);
	return jmin;
}

int main(int argc, char **argv)
{
	int i, j, arg;
	int cflag = 0;
	int nflag = 0;
	int Iflag = 0;
	double cdata = -1.0;
	Geometry *g;
	char *ifn = "-";
	FILE *fin;

	for(arg=1; arg<argc; ++arg)
	  if(argv[arg][0] == '-')
	    switch(argv[arg][1])
	    {
	      case 'c':
		cflag = 1;
		if(argv[arg][2] != '\0')
		  sscanf(argv[arg] + 2, "%lf", &cdata);
		break;

	      case 'n':
		nflag = 1;
		break;

	      case 'I':
		Iflag = 1;
		break;

	      default:
	      	fprintf(stderr, usage, argv[0]);
	      	exit(1);
	    }
	  else
	    ifn = argv[arg];

	fin = fopen_with_default(ifn, "r", NULL);

	while((g = Gread(fin)) != NULL)
	{
	  if(Iflag) {
	    double mom[3], xcam[3], axes[3*3];
	    double *mass = dmalloc(g->n);
	    for(i=0; i<g->n; ++i)
	      mass[i] = GuessElement(g->aname[i])->mass;
	    Gmoment(g, mass, mom, xcam, axes);
	    printf("com = {%g,%g,%g} I = {%g,%g,%g}\n",
	      xcam[0], xcam[1], xcam[2], mom[0], mom[1], mom[2]);
	    myfree(mass);
	  }

	  if(nflag && g->n > 1) {
	    int jmin;
	    double dmin;
	    for(i=0; i<(g->n); ++i)
	    {
	      jmin = nearest_neighbor(g, i, &dmin);
	      printf("%f atoms %d %d (%s-%s)\n",
		dmin, i + 1, jmin + 1, g->aname[i], g->aname[jmin]);
	    }
	  }

	  if(cflag) {
	    int nn;
	    double dmin, dmin2, dij[3];
	    Element *e;
	    for(i=0; i<(g->n); ++i) {
	      if(cdata > 0.0)
		dmin = cdata;
	      else {
		nearest_neighbor(g, i, &dmin);
		dmin *= 1.1;
	      }
	      dmin2 = dmin * dmin;
	      nn = 0;
	      for(j=0; j<(g->n); ++j) {
		if(j != i && Gdisp2(g, i, j, dij) <= dmin2) ++nn;
	      }
	      e = GuessElement(g->aname[i]);
	      printf("%d atom %d (%s) (within %f)\n",
		nn, i + 1, e->name, dmin);
	    }
	  }

	  Gfree(g);
	}
	if(fin != stdin) fclose(fin);

	return 0;
}
