/* GENETIC ALGORITHM AND RELATED CODE
 * Copyright (c) 1995 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: georand.c,v 1.7 1996/05/01 18:49:03 deaven Exp deaven $
 */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "dmd/geometry.h"
#include "dmd/random.h"
#include "dmd/memory.h"
#include "dmd/element.h"

#define AMU 1.66053e-24 /* grams per atomic mass unit */
#define CM  5.29177e-9  /* cm per bohr */

static char *usage = "usage: %s [options]\n"
  "\t-a name     set atom name (e.g. \"-a C\" for carbon)\n"
  "\t-n #        set number of atoms\n"
  "\t-m #        set number of geometries to create\n"
  "\t-c geo      use periodicity and box from given geometry\n"
  "\t-d #        scale box to set density (g/cc)\n"
  "\t-v #        scale box to set volume (bohr^3)\n"
  "\t-r #        set near neighbor distance minimum (bohr)\n"
  "\t-R #        set farthest group distance maximum (bohr)\n"
  "\t-f min,max  choose random flag values in a given range\n"
  "\t-s #        set random seed\n"
  "\t-t type     select a filter for the positions\n"
  "\t-o file     output file\n"
  ;

extern int ran3_seed;

static void getnew_box(double *x)
{
	x[0] = ran3() - 0.5;
	x[1] = ran3() - 0.5;
	x[2] = ran3() - 0.5;
}

static void getnew_sphere(double *x)
{
	double r;
	do {
	  getnew_box(x);
	  r = x[0] * x[0] + x[1] * x[1] + x[2] * x[2];
	} while(r > 0.25);
}

static void getnew_shell(double *x)
{
	double r;
	getnew_box(x);
	r = x[0] * x[0] + x[1] * x[1] + x[2] * x[2];
	r = 1.0 / sqrt(r);
	x[0] *= r;
	x[1] *= r;
	x[2] *= r;
}

static struct {
	char *name;
	void (*getnew)(double *);
} table[] = {
  {"box", getnew_box},
  {"sphere", getnew_sphere},
  {"shell", getnew_shell},
  {NULL, NULL}
};

void main(int argc, char **argv)
{
	int i, j, try, m = 1, n = 1;
	double dmin = -1.0, dmax = -1.0;
	double set_volume = -1.0, set_density = -1.0;
	Geometry *g = NULL;
	int fflag = 0, fmin, fmax;
	char buf[128];
	char *aname = "C";
	char *ofn = "-";
	void (*getnew)(double *) = table[0].getnew;
	FILE *fout;
	Element *e;

	ran3_seed = -1965;

	for(i=1; i<argc; ++i) {
	  if(argv[i][0] == '-') switch(argv[i][1]) {
	    case 'a':
	      aname = argv[++i];
	      break;

	    case 'o':
	      ofn = argv[++i];
	      break;

	    case 'c':
	      {
		FILE *f = fopen(argv[++i], "r");
		if(f == NULL) {
		  fprintf(stderr, "cannot open `%s'\n", argv[i]);
		  exit(1);
		}
		if((g = Gread(f)) == NULL) {
		  fprintf(stderr, "cannot read geometry from `%s'\n", argv[i]);
		  exit(1);
		}
		fclose(f);
	      }
	      break;

	    case 'v':
	      sscanf(argv[++i], "%lf", &set_volume);
	      break;

	    case 's':
	      sscanf(argv[++i], "%d", &ran3_seed);
	      if(ran3_seed > 0) ran3_seed *= -1;
	      break;

	    case 'f':
	      fflag = 1;
	      sscanf(argv[++i], "%d,%d", &fmin, &fmax);
	      break;

	    case 'd':
	      sscanf(argv[++i], "%lf", &set_density);
	      break;

	    case 'r':
	      sscanf(argv[++i], "%lf", &dmin);
	      break;

	    case 'R':
	      sscanf(argv[++i], "%lf", &dmax);
	      break;

	    case 'n':
	      sscanf(argv[++i], "%d", &n);
	      break;

	    case 'm':
	      sscanf(argv[++i], "%d", &m);
	      break;

	    case 't':
	      {
		char *name = argv[++i];
		for(j=0; table[j].name != NULL; ++j) {
		  if(strcmp(name, table[j].name) == 0) {
		    getnew = table[j].getnew;
		    break;
		  }
		}
		if(table[j].name == NULL) {
		  fprintf(stderr, "%s: unknown type `%s'\n", argv[0], name);
		  exit(1);
		}
	      }
	      break;

	    default:
	      fprintf(stderr, usage, argv[0]);
	      exit(1);
	  }
	}

	if((e = FindElement(aname)) == NULL) {
	  e = GuessElement(aname);
	  fprintf(stderr,
	    "cannot get info about element `%s', assuming it's `%s'\n",
	    aname, e->name);
	}

	sprintf(buf, "random %s%d", aname, n);

	if(g == NULL) {
	  double a;
	  g = Gmalloc(1, NULL);
	  a = (4.0/3.0 * M_PI) * n / Gvolume(g);
	  a = pow(a, 1.0/3.0) * (e->radius > 0.0 ? e->radius : 1.0);
	  for(i=0; i<9; ++i) g->box[i] *= a;
	}

	if(set_density > 0.0) 
	{
	  double a = (n * AMU * 
	    (e->mass > 0.0 ? e->mass : 1.0) / Gvolume(g)) / set_density;
	  a = pow(a, 1.0/3.0) / CM;
	  for(i=0; i<9; ++i) g->box[i] *= a;
	  sprintf(buf, "density=%.3f g/cc", set_density);
	}

	if(set_volume > 0.0) 
	{
	  double a = pow(set_volume / Gvolume(g), 1.0/3.0);
	  for(i=0; i<9; ++i) g->box[i] *= a;
	  sprintf(buf, "volume=%.3f bohr^3", set_volume);
	}
	else if(dmin > 0.0) 
	{
	  double a = dmin * pow(2.0 * (M_PI / 3.0) * n / Gvolume(g), 1.0/3.0);
	  for(i=0; i<9; ++i) g->box[i] *= a;
	}

	g = Grealloc(g, n);
	for(i=0; i<n; ++i) g->aname[i] = aname;

	fout = (strcmp(ofn, "-") == 0 ? stdout : fopen(ofn, "w"));
	if(fout == NULL) {
	  fprintf(stderr, "cannot open `%s'\n", ofn);
	  exit(1);
	}

	while(m-- > 0)
	{
	  int warn = 64;
	  for(i=try=0; i<n; ++i)
	  {
	    getnew(g->x + 3*i);
	    if(dmin > 0.0) {
	      for(j=0; j<i; ++j)
		if(Gdist(g, i, j) < dmin)
		{
		  if(++try > warn) {
		    if(warn < 2*n) { /* give up after this */
		      fprintf(stderr,
			"having difficulty fitting "
			"in all atoms (%d failed attempts).\n", try);
		      warn *= 2;
		    }
		  }
		  else
		  {
		    --i;
		  }
		  break;
		}
	    }
	  }
	  if(dmin > 0.0) Gseparate(g, dmin); /* fix up, maybe */
	  if(dmax > 0.0) Gconnect(g, dmax);
	  g->name = buf;

	  if(fflag) {
	    g->flags = myrealloc(g->flags, g->n * sizeof(int));
	    for(i=0; i<g->n; ++i)
	      g->flags[i] = fmin + iran3(fmax - fmin + 1);
	  }

	  Gprint(fout, g);
	}

	if(fout != stdout) fclose(fout);

	exit(0);
}
