/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: geordf.c,v 1.3 1995/11/01 23:44:27 deaven Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "dmd/geometry.h"
#include "dmd/histogram.h"
#include "dmd/memory.h"
#include "dmd/tools.h"

char *usage = "usage: %s [options] [geometry]\n"
  "options:\n"
  "\t-d rc            set cutoff distance\n"
  "\t-r res           set distance/angular resolution\n"
  "\t-o fn            set output filename\n"
  "\t-m fn            select atoms in file `fn'\n"
  "\t-a               make bond-angular distribution\n"
  ;

struct neighbor {
	double r;
	double dx[3];
};

/* Given two atoms i and j in a geometry g, determine the distance
 * and box displacement between the atoms.
 *
 * This routine is the bottleneck in this code, so it must be kept
 * efficient.
 */
static struct neighbor *dist(Geometry *g, int i, int j, int ncell[3])
{
        register int m, m3;
        register double dx;
	static struct neighbor nei;
        for(m=0; m<3; ++m)
          nei.dx[m] = g->x[j*3+m] - g->x[i*3+m] + ncell[m];
        nei.r = 0.0;
        for(m=0; m<3; ++m) {
	  m3 = m * 3;
          dx = nei.dx[0] * g->box[m3    ] +
	       nei.dx[1] * g->box[m3 + 1] +
	       nei.dx[2] * g->box[m3 + 2];
          nei.r += dx * dx;
        }
        return &nei;
}

static double boxlen(Geometry *g, int m)
{
	double d, *b = g->box;
	d = b[m] * b[m] + b[m + 3] * b[m + 3] + b[m + 6] * b[m + 6];
	return sqrt(d);
}

static double angle(struct neighbor *a, struct neighbor *b)
{
	double dot = 
	  a->dx[0] * b->dx[0] +
	  a->dx[1] * b->dx[1] +
	  a->dx[2] * b->dx[2];
	dot /= (a->r * b->r);
	return acos(dot);
}

/* Given a list of neighbors, find the minimum angle formed by
 * neighbor j and any other neighbor.
 */
static double minangle(struct neighbor *neitab, int nnei, int j)
{
	int k;
	double a, result = 2.0 * M_PI;
	for(k=0; k<nnei; ++k) {
	  if(k != j) {
	    a = angle(neitab + j, neitab + k);
	    if(a < result) result = a;
	  }
	}
	return result;
}

/* Read a geometry and write a radial distribution histogram.
 */
void main(int argc, char **argv)
{
	int i, j, arg, nmask, nangle, nnei = 0;
	int ncell[3], nmax[3];
	double rc2;
	Geometry *g;
	GeometryMask *mask = NULL;
	Histogram *H;
	struct neighbor *nei, *neitab = NULL;
	FILE *f;
	double rc = -1.0;		/* cutoff distance */
	double dr = 0.1;		/* resolution (bohr) */
	int aflag = 0;			/* do angles, not distance */
	char *fn = "-", *ofn = "-";	/* filenames */

	for(arg=1; arg<argc; ++arg) {
	  if(argv[arg][0] == '-') {
	    switch(argv[arg][1]) {
	      case 'a':
		aflag = 1;
		break;
	      case 'd':
		sscanf(argv[++arg], "%lf", &rc);
		break;
	      case 'r':
		sscanf(argv[++arg], "%lf", &dr);
		break;
	      case 'o':
		ofn = argv[++arg];
		break;
	      case 'm':
		{
		  GeometryMask *and;
		  f = fopen_with_default(argv[++arg], "r", NULL);
		  if((and = GMaskRead(f)) == NULL) {
		    fprintf(stderr, "%s: cannot read mask `%s'\n", 
		      argv[0], argv[arg]);
		    exit(1);
		  }
		  fclose(f);
		  mask = GMaskOr(mask, and);
		  GMaskDestroy(and);
		}
		break;
	      default:
		fprintf(stderr, usage, argv[0]);
		exit(1);
	    }
	  } else { /* geometry filename */
	    fn = argv[arg];
	  }
	}

	f = fopen_with_default(fn, "r", NULL);
	g = Gread(f);
	fclose(f);

	if(g == NULL) {
	  fprintf(stderr, "%s: cannot read geometry\n", argv[0]);
	  exit(1);
	}

	if(mask == NULL) { /* has not bbeen set */
	  mask = GMask(g->n);
	  for(i=0; i<g->n; ++i) mask->s[i] = 1;
	}

	if(rc < 0.0) {	/* has not been set */
	  rc = 3.0 * pow(Gvolume(g), 1.0 / 3.0);
	}
	rc2 = rc * rc;

	for(i=0; i<3; ++i) {
	  if(g->period & GP_PDIR(i))
	    nmax[i] = 1.1 * rc / boxlen(g, i) + 2;
	  else
	    nmax[i] = 0;
	}

	H = Histmalloc(dr);
	nmask = nangle = 0;
	for(i=0; i<g->n; ++i) {
	  if(GMaskIsSelected(mask, i)) {
	    myfree(neitab);
	    neitab = NULL;
	    nnei = 0;
	    for(j=0; j<g->n; ++j) {
	      for(ncell[0] = -nmax[0]; ncell[0] <= nmax[0]; ++ncell[0])
	      for(ncell[1] = -nmax[1]; ncell[1] <= nmax[1]; ++ncell[1])
	      for(ncell[2] = -nmax[2]; ncell[2] <= nmax[2]; ++ncell[2])
	      if(i != j || ncell[0] != 0 || ncell[1] != 0 || ncell[2] != 0)
	      {
		nei = dist(g, i, j, ncell);
		if(nei->r <= rc2) {
		  nei->r = sqrt(nei->r);
		  if(aflag) {
		    neitab = myrealloc(neitab, (nnei + 1) * sizeof(*neitab));
		    neitab[nnei].r = nei->r;
		    mmul3(g->box, nei->dx, neitab[nnei].dx);
		    ++nnei;
		  }
		  else
		    Histadd(H, nei->r, 1.0 / (nei->r * nei->r));
		}
	      } /* if, ncell[] */
	    } /* j */
	    if(aflag && nnei > 2) {
	      for(j=0; j<nnei; ++j) {
		Histadd(H, (180.0 / M_PI) * minangle(neitab, nnei, j), 1.0);
		++nangle;
	      }
	    }
	    ++nmask;
	  }
	}
	if(nmask == 0)
	  fprintf(stderr, "%s: warning: no atoms were selected\n", argv[0]);
	else {
	  fprintf(stderr, "%s: selected %d/%d atoms\n", argv[0], nmask, g->n);
	  if(aflag && nangle == 0)
	    fprintf(stderr, "%s: warning: no angles were selected\n", argv[0]);
	  else {
	    for(i=0; i<H->n; ++i)
	      H->data[i].z *= (aflag ? 1.0 / nangle : (4.0 * M_PI) / nmask);
	  }
	}

	f = fopen_with_default(ofn, "w", stdout);
	if(aflag)
	  fprintf(f, "2 #1=(theta [deg]) #2=(n(theta) %s)\n", g->name);
	else
	  fprintf(f, "2 #1=(r [bohr]) #2=(g(r) %s)\n", g->name);
	HistXY(f, H);
	if(f != stdout) fclose(f);
	Histfree(H);
	GMaskDestroy(mask);
	Gfree(g);

	exit(0);
}
